import inspect
from pathlib import Path
from typing import List, Optional

from RPA.core.windows.locators import Locator

from RPA.Windows import utils
from RPA.Windows.keywords import (
    ActionNotPossible,
    keyword,
    LibraryContext,
)

if utils.IS_WINDOWS:
    import uiautomation as auto
    from uiautomation import TreeNode


class ElementKeywords(LibraryContext):
    """Keywords for handling Control elements"""

    @keyword
    def set_anchor(
        self,
        locator: Locator,
        timeout: Optional[float] = None,
    ) -> None:
        """Set anchor to an element specified by the locator.

        All following keywords using locators will use this element
        as an root element. Specific use case could be setting
        anchor to TableControl element and then getting column data
        belonging to that TableControl element.

        To release anchor call ``Clear Anchor`` keyword.

        :param locator: string locator or Control element
        :param timeout: timeout in seconds for element lookup (default 10.0)

        Example:

        .. code-block:: robotframework

            Set Anchor  type:Table name:Orders depth:16
            FOR  ${row}  IN RANGE  200
                ${number}=  Get Value   name:number row ${row}
                Exit For Loop If   $number == ${EMPTY}
                ${sum}=  Get Value   name:sum row ${row}
                Log   Order number:${number} has sum:{sum}
            END
            Clear Anchor
        """
        self.ctx.anchor_element = self.ctx.get_element(locator, timeout=timeout)

    @keyword
    def clear_anchor(self) -> None:
        """Clears control anchor set by ``Set Anchor``

        This means that all following keywords accessing elements
        will use active window or desktop as root element.
        """
        self.ctx.anchor_element = None

    @keyword
    def print_tree(
        self,
        locator: Optional[Locator] = None,
        max_depth: int = 8,
        capture_image_folder: Optional[str] = None,
        log_as_warnings: bool = False,
    ) -> None:
        """Print Control element tree.

        Windows application structure can contain multilevel element
        structure. Understanding this structure is important for
        creating locators.

        This keyword can be used to output application element structure
        starting with the element defined by the `locator`.

        :param locator: string locator or Control element
        :param max_depth: maximum depth level (defaults to 8)
        :param capture_image_folder: if None images are not captured
        :param log_as_warnings: if set log messages are visible on the console
        """

        def GetFirstChild(ctrl: TreeNode) -> TreeNode:
            return ctrl.GetFirstChildControl()

        def GetNextSibling(ctrl: TreeNode) -> TreeNode:
            return ctrl.GetNextSiblingControl()

        index = 1
        target_elem = self.ctx.get_element(locator)
        image_folder = None
        if capture_image_folder:
            image_folder = Path(capture_image_folder).expanduser().resolve()
            image_folder.mkdir(parents=True, exist_ok=True)
        control_log = self.logger.warning if log_as_warnings else self.logger.info

        for control, depth in auto.WalkTree(
            target_elem.item,
            getFirstChild=GetFirstChild,
            getNextSibling=GetNextSibling,
            includeTop=True,
            maxDepth=max_depth,
        ):
            control_str = str(control)
            if image_folder:
                capture_filename = f"{control.ControlType}_{index}.png"
                img_path = str(image_folder / capture_filename)
                try:
                    control.CaptureToImage(img_path)
                except Exception as exc:  # pylint: disable=broad-except
                    self.logger.warning(
                        "Couldn't capture into %r due to: %s", img_path, exc
                    )
                else:
                    control_str += f" [{capture_filename}]"
            control_log(" " * depth * 4 + control_str)
            index += 1

    @keyword
    def get_attribute(self, locator: Locator, attribute: str) -> str:
        """Get attribute value of the element defined by the locator.

        :param locator: string locator or Control element
        :param attribute: name of the attribute to get
        :return: value of attribute

        Example:

        .. code-block:: robotframework

            ${id}=   Get Attribute  type:Edit name:firstname   AutomationId
        """
        # TODO. Add examples
        element = self.ctx.get_element(locator)
        attr = hasattr(element.item, attribute)
        if not attr:
            raise ActionNotPossible(
                f"Element found with {locator!r} does not have {attribute!r} attribute"
            )
        if callable(attr):
            raise ActionNotPossible(
                f"Can't access attribute {attribute!r} of element {element!r}"
            )
        return str(getattr(element.item, attribute))

    @keyword
    def list_attributes(self, locator: Locator) -> List:
        """List all element attributes.

        :param locator: string locator or Control element
        :return: list of element attributes (strings)
        """
        element = self.ctx.get_element(locator)
        element_attributes = [e for e in dir(element.item) if not e.startswith("_")]
        attributes = []

        for attr_name in element_attributes:
            attr = getattr(element.item, attr_name)
            if not inspect.ismethod(attr):
                attributes.append(attr_name)
        return attributes



from pathlib import Path
from typing import Optional

from RPA.core.windows.locators import Locator, WindowsElement

from RPA.Windows import utils
from RPA.Windows.keywords import keyword
from RPA.Windows.keywords.context import ActionNotPossible, LibraryContext

if utils.IS_WINDOWS:
    import uiautomation as auto


class ActionKeywords(LibraryContext):
    """Keywords for performing desktop actions"""

    @keyword(tags=["action", "mouse"])
    def click(
        self,
        locator: Locator,
        wait_time: Optional[float] = None,
        timeout: Optional[float] = None,
    ) -> WindowsElement:
        """Mouse click on element matching given locator.

        Exception ``ActionNotPossible`` is raised if element does not
        allow Click action.

        :param locator: string locator or Control element
        :param wait_time: time to wait after click, default is a
         library `wait_time`, see keyword ``Set Wait Time``
        :param timeout: float value in seconds, see keyword
         ``Set Global Timeout``
        :return: WindowsElement object

        Example:

        .. code-block:: robotframework

            Click  id:button1
            Click  id:button2 offset:10,10
            ${element}=  Click  name:SendButton  wait_time=5.0
        """
        return self._mouse_click(locator, "Click", wait_time, timeout)

    @keyword(tags=["action", "mouse"])
    def double_click(
        self,
        locator: Locator,
        wait_time: Optional[float] = None,
        timeout: Optional[float] = None,
    ) -> WindowsElement:
        """Double mouse click on element matching given locator.

        Exception ``ActionNotPossible`` is raised if element does not
        allow Click action.

        :param locator: string locator or Control element
        :param wait_time: time to wait after click, default is a
         library `wait_time`, see keyword ``Set Wait Time``
        :param timeout: float value in seconds, see keyword
         ``Set Global Timeout``
        :return: WindowsElement object

        Example:

        .. code-block:: robotframework

            ${element}=  Double Click  name:ResetButton
        """
        return self._mouse_click(locator, "DoubleClick", wait_time, timeout)

    @keyword(tags=["action", "mouse"])
    def right_click(
        self,
        locator: Locator,
        wait_time: Optional[float] = None,
        timeout: Optional[float] = None,
    ) -> WindowsElement:
        """Right mouse click on element matching given locator.

        Exception ``ActionNotPossible`` is raised if element does not
        allow Click action.

        :param locator: string locator or Control element
        :param wait_time: time to wait after click, default is a
         library `wait_time`, see keyword ``Set Wait Time``
        :param timeout: float value in seconds, see keyword
         ``Set Global Timeout``
        :return: WindowsElement object

        Example:

        .. code-block:: robotframework

            ${element}=  Right Click  name:MenuButton
        """
        return self._mouse_click(locator, "RightClick", wait_time, timeout)

    @keyword(tags=["action", "mouse"])
    def middle_click(
        self,
        locator: Locator,
        wait_time: Optional[float] = None,
        timeout: Optional[float] = None,
    ) -> WindowsElement:
        """Right mouse click on element matching given locator.

        Exception ``ActionNotPossible`` is raised if element does not
        allow Click action.

        :param locator: string locator or Control element
        :param wait_time: time to wait after click, default is a
         library `wait_time`, see keyword ``Set Wait Time``
        :param timeout: float value in seconds, see keyword
         ``Set Global Timeout``
        :return: WindowsElement object

        Example:

        .. code-block:: robotframework

            ${element}=  Middle Click  name:button2
        """
        return self._mouse_click(locator, "MiddleClick", wait_time, timeout)

    def _mouse_click(
        self,
        locator: Locator,
        click_type: str,
        wait_time: Optional[float],
        timeout: Optional[float],
    ) -> WindowsElement:
        click_wait_time: float = wait_time or self.ctx.wait_time
        with self.set_timeout(timeout):
            element = self.ctx.get_element(locator)
            self._click_element(element, click_type, click_wait_time)
        return element

    def _click_element(
        self, element: WindowsElement, click_type: str, click_wait_time: float
    ):
        item = element.item
        click_function = getattr(item, click_type, None)
        if not click_function:
            raise ActionNotPossible(
                f"Element {element!r} does not have {click_type!r} attribute"
            )

        # Attribute added in `RPA.core.windows.locators.LocatorMethods`.
        offset: Optional[str] = item.robocorp_click_offset
        offset_x: Optional[int] = None
        offset_y: Optional[int] = None
        log_message = f"{click_type}-ing element"
        if offset:
            # Get a new fresh bounding box each time, since the element might have been
            #  moved from its initial spot.
            rect = item.BoundingRectangle
            # Now compute the new coordinates starting from the element center.
            dist_x, dist_y = (int(dist.strip()) for dist in offset.split(","))
            pos_x, pos_y = rect.xcenter() + dist_x, rect.ycenter() + dist_y
            # If we pass the newly obtained absolute position to the clicking function
            #  that won't work as expected. You see (`help(item.Click)`), if the passed
            #  offset is positive then it gets relative to the left-top corner and if
            #  is negative then the right-bottom corner is used.
            # Let's assume we end up with a positive relative offset. (using left-top
            #  fixed corner)
            offset_x, offset_y = pos_x - rect.left, pos_y - rect.top
            # If by any chance an offset is negative, it gets relative to the
            #  right-bottom corner, therefore adjust it accordingly.
            if offset_x < 0:
                offset_x -= rect.width()
            if offset_y < 0:
                offset_y -= rect.height()
            log_message += f" with offset: {offset_x}, {offset_y}"

        self.logger.debug(log_message)
        click_function(
            x=offset_x,
            y=offset_y,
            simulateMove=self.ctx.SIMULATE_MOVE,
            waitTime=click_wait_time,
        )

    @keyword(tags=["action"])
    def select(self, locator: Locator, value: str) -> WindowsElement:
        """Select value on Control element if action is supported.

        Exception ``ActionNotPossible`` is raised if element does not
        allow Select action.

        :param locator: string locator or Control element
        :param value: string value to select on Control element
        :return: WindowsElement object

        Example:

        .. code-block:: robotframework

            Select  type:SelectControl   option2
        """
        element = self.ctx.get_element(locator)
        if hasattr(element.item, "Select"):
            element.item.Select(value)
        else:
            raise ActionNotPossible(
                f"Element {locator!r} does not have 'Select' attribute"
            )
        return element

    @keyword(tags=["action", "keyboard"])
    def send_keys(
        self,
        locator: Optional[Locator] = None,
        keys: Optional[str] = None,
        interval: float = 0.01,
        wait_time: Optional[float] = None,
        send_enter: bool = False,
    ):
        """Send keys to desktop, current window or to Control element
        defined by given locator.

        If ``locator`` is `None` then keys are sent to desktop.

        Exception ``ActionNotPossible`` is raised if element does not
        allow SendKeys action.

        :param locator: string locator or Control element (default None means desktop)
        :param keys: the keys to send
        :param interval: time between sending keys, default 0.01 seconds
        :param wait_time: time to wait after sending keys, default is a
         library `wait_time`, see keyword ``Set Wait Time``
        :param send_enter: if True then {Enter} is sent at the end of the keys
        :return: WindowsElement object

        Example:

        .. code-block:: robotframework

            Send Keys  desktop   {Ctrl}{F4}
            Send Keys  keys={Ctrl}{F4}   # locator will be NONE, keys send to desktop
            ${element}=   Send Keys  id:input5  username   send_enter=True
            ${element}=   Get Element   id:pass
            Send Keys  ${element}  password   send_enter=True
        """
        if locator:
            element = self.ctx.get_element(locator).item
        else:
            element = auto
        keys_wait_time = wait_time or self.ctx.wait_time
        if send_enter:
            keys += "{Enter}"
        if hasattr(element, "SendKeys"):
            self.logger.info("Sending keys %r to element %r", keys, element)
            element.SendKeys(text=keys, interval=interval, waitTime=keys_wait_time)
        else:
            raise ActionNotPossible(
                f"Element found with {locator!r} does not have 'SendKeys' attribute"
            )

    @keyword(tags=["action"])
    def get_text(self, locator: Locator) -> str:
        """Get text from Control element defined by the locator.

        Exception ``ActionNotPossible`` is raised if element does not
        allow GetWindowText action.

        :param locator: string locator or Control element
        :return: value of WindowText attribute of an element

        Example:

        .. code-block:: robotframework

            ${date} =  Get Text   type:Edit name:"Date of birth"
        """
        element = self.ctx.get_element(locator)
        if hasattr(element.item, "GetWindowText"):
            return element.item.GetWindowText()
        raise ActionNotPossible(
            f"Element found with {locator!r} does not have 'GetWindowText' attribute"
        )

    @keyword(tags=["action"])
    def get_value(self, locator: Locator) -> Optional[str]:
        """Get value of the element defined by the locator.

        Exception ``ActionNotPossible`` is raised if element does not
        allow GetValuePattern action.

        :param locator: string locator or Control element
        :return: value of ValuePattern attribute of an element

        Example:

        .. code-block:: robotframework

            ${value}=   Get Value   type:DataItem name:column1
        """
        element = self.ctx.get_element(locator)
        if hasattr(element.item, "GetValuePattern"):
            value_pattern = element.item.GetValuePattern()
            return value_pattern.Value if value_pattern else None

        raise ActionNotPossible(
            f"Element found with {locator!r} does not have 'GetValuePattern' attribute"
        )

    @keyword(tags=["action"])
    def set_value(
        self,
        locator: Optional[Locator] = None,
        value: Optional[str] = None,
        append: bool = False,
        enter: bool = False,
        newline: bool = False,
    ) -> WindowsElement:
        """Set value of the element defined by the locator.

        *Note.* Anchor works only on element structures where it can
        be relied on that root/child element tree will remain the same.
        Usually these kind of structures are tables.

        Exception ``ActionNotPossible`` is raised if element does not
        allow SetValue action.

        :param locator: string locator or Control element
        :param value: string value to set
        :param append: False for setting value, True for appending value
        :param enter: set True to press enter key at the end of the line
        :param newline: set True to add newline to the end of value
        :return: WindowsElement object

        *Note.* It is important to set ``append=True`` if you want keep text in
        the element. Other option is to read current text into a variable and
        modify that value to pass for ``Set Value`` keyword.

        Example:

        .. code-block:: robotframework

            Set Value   type:DataItem name:column1   ab c  # Set value to "ab c"
            # Press ENTER after setting the value
            Set Value    type:Edit name:"File name:"    console.txt    enter=True

            # Add newline (manually) at the end of the string (Notepad example)
            Set Value    name:"Text Editor"  abc\\n
            # Add newline with parameter
            Set Value    name:"Text Editor"  abc   newline=${True}

            # Clear Notepad window and start appending text
            Set Anchor  name:"Text Editor"
            # all following keyword calls will use anchor element as locator
            # UNLESS they specify locator specifically or `Clear Anchor` is used
            ${time}=    Get Time
            # Clears when append=False (default)
            Set Value    value=time now is ${time}
            # Append text and add newline to the end
            Set Value    value= and it's task run time    append=True    newline=True
            # Continue appending
            Set Value    value=this will appear on the 2nd line    append=True
        """
        value = value or ""
        element = self.ctx.get_element(locator)
        current_value = ""
        newline_string = "\n" if newline else ""
        if hasattr(element.item, "GetValuePattern"):
            value_pattern = element.item.GetValuePattern()
            if append:
                current_value = value_pattern.Value
            value_pattern.SetValue(f"{current_value}{value}{newline_string}")
        elif hasattr(element.item, "GetLegacyIAccessiblePattern"):
            pattern = element.item.GetLegacyIAccessiblePattern()
            if append:
                current_value = pattern.Value
            pattern.SetValue(f"{current_value}{value}{newline_string}")
        else:
            raise ActionNotPossible(
                f"Element found with {locator!r} doesn't support value setting"
            )
        if enter:
            self.send_keys(element, "{Ctrl}{End}{Enter}")
        return element

    @keyword(tags=["action"])
    def set_wait_time(self, wait_time: float) -> float:
        """Set library wait time for action keywords.

        The wait_time is spent after each keyword performing
        mouse or keyboard action.

        Library default wait_time is `0.5`

        Returns value of the previous wait_time value.

        :param wait_time: float value (in seconds), e.g. `0.1`
        :return: previous wait value

        Example:

        .. code-block:: robotframework

            ${old_wait_time}=  Set Wait Time  0.2
        """
        old_value = self.ctx.wait_time
        self.ctx.wait_time = wait_time
        return old_value

    @keyword(tags=["action"])
    def screenshot(self, locator: Locator, filename: str) -> str:
        """Take a screenshot of the element defined by the locator.

        Exception ``ActionNotPossible`` is raised if element does not
        allow CaptureToImage action.

        :param locator: string locator or Control element
        :param filename: image filename
        :return: absolute path to the screenshot file

        Example:

        .. code-block:: robotframework

            Screenshot  desktop   desktop.png
            Screenshot  subname:Notepad   notepad.png
        """
        element = self.ctx.get_element(locator)
        if not hasattr(element.item, "CaptureToImage"):
            raise ActionNotPossible(
                f"Element found with {locator!r} does not have 'CaptureToImage' "
                "attribute"
            )

        element.item.SetFocus()
        filepath = str(Path(filename).expanduser().resolve())
        element.item.CaptureToImage(filepath)
        return filepath

    @keyword(tags=["action"])
    def set_global_timeout(self, timeout: float) -> float:
        """Set global timeout for element search. Applies also
        to ``Control Window`` keyword.

        By default library has timeout of 10 seconds.

        :param timeout: float value in seconds
        :return: previous timeout value

        Example:

        .. code-block:: robotframework

            ${old_timeout}=  Set Global Timeout  20
            ${old_timeout}=  Set Global Timeout  9.5
        """
        previous_timeout = self.ctx.global_timeout
        self.ctx.global_timeout = timeout
        auto.SetGlobalSearchTimeout(self.ctx.global_timeout)
        return previous_timeout

    @keyword(tags=["action"])
    def set_focus(self, locator: Locator) -> None:
        """Set view focus to the element defined by the locator.

        :param locator: string locator or Control element

        Example:

        .. code-block:: robotframework

            Set Focus  name:Buy type:Button
        """
        element = self.ctx.get_element(locator)
        if not hasattr(element.item, "SetFocus"):
            raise ActionNotPossible(
                f"Element found with {locator!r} does not have 'SetFocus' attribute"
            )
        element.item.SetFocus()

    @keyword(tags=["action", "mouse"])
    def drag_and_drop(
        self,
        source_element: Locator,
        target_element: Locator,
        speed: Optional[float] = 1.0,
        copy: Optional[bool] = False,
        wait_time: Optional[float] = 1.0,
    ):
        """Drag and drop the source element into target element.

        :param source: source element for the operation
        :param target: target element for the operation
        :param speed: adjust speed of operation, bigger value means more speed
        :param copy: on True does copy drag and drop, defaults to move
        :param wait_time: time to wait after drop, default 1.0 seconds

        Example:

        .. code-block:: robotframework

            # copying a file, report.html, from source (File Explorer) window
            # into a target (File Explorer) Window
            # locator
            Drag And Drop
            ...    name:C:\\temp type:Windows > name:report.html type:ListItem
            ...    name:%{USERPROFILE}\\Documents\\artifacts type:Windows > name:"Items View"
            ...    copy=True

        Example:

        .. code-block:: robotframework

            # moving *.txt files into subfolder within one (File Explorer) window
            ${source_dir}=    Set Variable    %{USERPROFILE}\\Documents\\test
            Control Window    name:${source_dir}
            ${files}=    Find Files    ${source_dir}${/}*.txt
            # first copy files to folder2
            FOR    ${file}    IN    @{files}
                Drag And Drop    name:${file.name}    name:folder2 type:ListItem    copy=True
            END
            # second move files to folder1
            FOR    ${file}    IN    @{files}
                Drag And Drop    name:${file.name}    name:folder1 type:ListItem
            END
        """  # noqa: E501
        source = self.ctx.get_element(source_element)
        target = self.ctx.get_element(target_element)
        try:
            if copy:
                auto.PressKey(auto.Keys.VK_CONTROL)
            auto.DragDrop(
                source.xcenter,
                source.ycenter,
                target.xcenter,
                target.ycenter,
                moveSpeed=speed,
                waitTime=wait_time,
            )
        finally:
            if copy:
                self.click(source)
                auto.ReleaseKey(auto.Keys.VK_CONTROL)



# pylint: disable=unused-import
from RPA.core.windows.context import (  # noqa: F401
    ElementNotFound,
    WindowControlError,
    WindowsContext as LibraryContext,
    with_timeout,
)


class ActionNotPossible(ValueError):
    """Action is not possible for the given Control."""


# Not yet used exceptions:


class ControlNotFound(ValueError):
    """No matching controls were found."""


class MultipleControlsFound(ValueError):
    """Multiple matching controls were found, but only one was expected."""


class TimeoutException(ValueError):
    """Timeout reached while waiting for condition."""



import logging
from typing import Optional

# pylint: disable=wrong-import-order
from RPA.core.windows import WindowsElementsMixin
from robotlibcore import DynamicCore

from . import utils
from .keywords import (
    ActionKeywords,
    ElementKeywords,
    LocatorKeywords,
    WindowKeywords,
)

if utils.IS_WINDOWS:
    # Configure comtypes to not generate DLL bindings into
    # current environment, instead keeping them in memory.
    # Slower, but prevents dirtying environments.
    import comtypes.client
    from uiautomation.uiautomation import Logger

    comtypes.client.gen_dir = None


# NOTE(cmiN): We use as base the robotframework `DynamicCore` this time instead of the
#  vendorized one, like found in `RPA.core.windows.WindowsElements`.
class Windows(WindowsElementsMixin, DynamicCore):
    # pylint: disable=anomalous-backslash-in-string
    """The `Windows` is a library that can be used for Windows desktop automation.

    **Installation**

    This library, ``RPA.Windows`` is available via **rpaframework-windows** package. This
    package is first available as a separate package installation from **rpaframework** package.
    In the next stage this is integrated into **rpaframework** package. Ultimately this library
    will replace current library ``RPA.Desktop.Windows`` if seen as approriate.

    .. code-block:: yaml

        channels:
          - conda-forge
        dependencies:
          - python=3.7.5
          - pip=20.1
          - pip:
            - rpaframework-windows==2.0.0


    **About terminology**

    **ControlType** is a value referred to by locator keys `type:` or `control`. Represents type of application
    object, which can be e.g. `Window`, `Button` or `ListItem`.

    **Element** is an entity of an application structure (e.g. certain button in a window), which can be
    identified by a locator.

    **WindowsElement** is an library container object for the ``Element``. All keywords returning element, return
    element as ``WindowsElement`` and all keywords taking in ``locator`` or ``root_element`` arguments accept
    ``WindowsElement`` as an argument value.

    Structure of the ``WindowsElement``

    .. code-block:: python

        class WindowsElement:
            item: Control        # ``item`` contains object instance of the element
            locator: str         # ``locator`` that found this element
            name: str            # ``Name`` attribute of the element
            automation_id: str   # ``AutomationId`` attribute of the element
            control_type: str    # ``ControlTypeName`` attribute of the element
            class_name: str      # ``ClassName`` attribute of the element
            left: int            # element's rectangle left coordinate
            right: int           # element's rectangle right coordinate
            top: int             # element's rectangle top coordinate
            bottom: int          # element's rectangle bottom coordinate
            width: int           # element's rectangle horizontal width
            height: int          # element's rectangle vertical height
            xcenter: int         # element's rectangle center point x coordinate
            ycenter: int         # element's rectangle center point y coordinate

    Example of the ``WindowsElement`` usage

    .. code-block:: robotframework

        ${rows}=    Get Elements    class:DataGridRow
        # ${rows} is a list of ``WindowsElement``s
        FOR    ${row}    IN    @{rows}
            Log To Console   ${row.name}                # access ``WindowsElement``
            Log To Console   ${row.item.AutomationId}   # access ``WindowsElement.item`` directly
            Log To Console   ${row.item.Name}           # same as ``${row.name}``
        END


    **Locators**

    Locators are based on different strategies that can used identify Control object.

    Available strategies that can be used for the locators:

    =============== =======================
    Key             Maps to search property
    =============== =======================
    name            Name
    class           ClassName
    type            ControlType
    control         ControlType
    id              AutomationId
    automationid    AutomationId
    regex           RegexName
    subname         SubName
    index           foundIndex (int)
    offset          offset coordinates (x (int), y (int)) from control center
    executable      target window by its executable name
    handle          target window handle (int)
    desktop         *SPECIAL* target desktop, no value for the key e.g. `desktop:desktop and name:Calculator`
    process         *NOT YET SUPPORTED* target window by its executable's process id
    depth           searchDepth (int) for finding Control (default 8)
    =============== =======================

    **About root element on locators**

    Locators work on currently active `root element`. At the start `root element` is the whole
    desktop. There are different ways on changing this root element.

    Keyword ``Control Window`` is the most common method of setting certain system window
    as a root element for further actions using locators.

    Locators themselves support cascading syntax (denoted by character `>` in the locator string),
    which can denote root element in "parent (root) & child" terms.

    For example.

    .. code-block:: robotframework

        Click  id:controls > id:activate

    On the above example the left side of the `>` character, `id:controls`, represents the root element
    (can be called as "parent element" in this case). Right side of the locator string, `id:activate`,
    represents "child" element and it will be searched under the "parent element".

    This way element search is more efficient, because search are restricted to certain section of element
    tree which can be quite huge especially on the desktop level and in certain applications.

    Keyword examples:

    .. code-block:: robotframework

        Control Window    name:Calculator
        Control Window    Calculator  # will execute search by 'name:Calculator'
        Control Window    executable:Spotify.exe

    some example locators, `and` can be omitted ie. space ` ` between locator keys means the same thing as `and`:

    .. code-block:: bash

        id:clearButton
        type:Group and name:"Number pad" > type:Button and index:4
        type:Group and name:"Number pad" > control:Button index:5
        id:Units1 > name:${unit}
        class:Button offset:370,0

    **About locator restrictions**

    Visual locators are not supported in this library and they can't be used in the same chain with these
    Windows locators. Visual locators are supported by the `RPA.Desktop` library. Locator chaining (image and
    Windows locators) support will be added in the future.

    Locator syntax does not yet support OR operation (only AND operations).

    **About search depth**

    The library does element search depth by default to the level of 8. This means that locator will look into
    8 levels of elements under element tree of the root element. This can lead into situation where element
    can't be found. To fix this it is recommended to set root element which can be found within 8 levels OR
    defining `depth` in the locator string to a bigger value, e.g. `id:deeplyNestedButton depth:16`. Useful
    keywords for setting root element are ``Control Window``, ``Set Anchor`` and ``Get Element``.

    **Keyboard and mouse**

    Keys for the keyboard actions are given using ``uiautomation`` specification.

    Special keys which are given within `{}` syntax.

    =================== =======================
    Key                 Maps to action
    =================== =======================
    LBUTTON             Left mouse button
    RBUTTON             Right mouse button
    CANCEL              Control-break processing
    MBUTTON             Middle mouse button (three-button mouse)
    XBUTTON1            X1 mouse button
    XBUTTON2            X2 mouse button
    BACK                BACKSPACE key
    TAB                 TAB key
    CLEAR               CLEAR key
    RETURN              ENTER key
    ENTER               ENTER key
    SHIFT               SHIFT key
    CTRL                CTRL key
    CONTROL             CTRL key
    ALT                 ALT key
    PAUSE               PAUSE key
    CAPITAL             CAPS LOCK key
    KANA                IME Kana mode
    HANGUEL             IME Hanguel mode (maintained for compatibility; use VK_HANGUL)
    HANGUL              IME Hangul mode
    JUNJA               IME Junja mode
    FINAL               IME final mode
    HANJA               IME Hanja mode
    KANJI               IME Kanji mode
    ESC                 ESC key
    ESCAPE              ESC key
    CONVERT             IME convert
    NONCONVERT          IME nonconvert
    ACCEPT              IME accept
    MODECHANGE          IME mode change request
    SPACE               SPACEBAR
    PRIOR               PAGE UP key
    PAGEUP              PAGE UP key
    NEXT                PAGE DOWN key
    PAGEDOWN            PAGE DOWN key
    END                 END key
    HOME                HOME key
    LEFT                LEFT ARROW key
    UP                  UP ARROW key
    RIGHT               RIGHT ARROW key
    DOWN                DOWN ARROW key
    SELECT              SELECT key
    PRINT               PRINT key
    EXECUTE             EXECUTE key
    SNAPSHOT            PRINT SCREEN key
    PRINTSCREEN         PRINT SCREEN key
    INSERT              INS key
    INS                 INS key
    DELETE              DEL key
    DEL                 DEL key
    HELP                HELP key
    WIN                 Left Windows key (Natural keyboard)
    LWIN                Left Windows key (Natural keyboard)
    RWIN                Right Windows key (Natural keyboard)
    APPS                Applications key (Natural keyboard)
    SLEEP               Computer Sleep key
    NUMPAD0             Numeric keypad 0 key
    NUMPAD1             Numeric keypad 1 key
    NUMPAD2             Numeric keypad 2 key
    NUMPAD3             Numeric keypad 3 key
    NUMPAD4             Numeric keypad 4 key
    NUMPAD5             Numeric keypad 5 key
    NUMPAD6             Numeric keypad 6 key
    NUMPAD7             Numeric keypad 7 key
    NUMPAD8             Numeric keypad 8 key
    NUMPAD9             Numeric keypad 9 key
    MULTIPLY            Multiply key
    ADD                 Add key
    SEPARATOR           Separator key
    SUBTRACT            Subtract key
    DECIMAL             Decimal key
    DIVIDE              Divide key
    F1                  F1 key
    F2                  F2 key
    F3                  F3 key
    F4                  F4 key
    F5                  F5 key
    F6                  F6 key
    F7                  F7 key
    F8                  F8 key
    F9                  F9 key
    F10                 F10 key
    F11                 F11 key
    F12                 F12 key
    F13                 F13 key
    F14                 F14 key
    F15                 F15 key
    F16                 F16 key
    F17                 F17 key
    F18                 F18 key
    F19                 F19 key
    F20                 F20 key
    F21                 F21 key
    F22                 F22 key
    F23                 F23 key
    F24                 F24 key
    NUMLOCK             NUM LOCK key
    SCROLL              SCROLL LOCK key
    LSHIFT              Left SHIFT key
    RSHIFT              Right SHIFT key
    LCONTROL            Left CONTROL key
    LCTRL               Left CONTROL key
    RCONTROL            Right CONTROL key
    RCTRL               Right CONTROL key
    LALT                Left MENU key
    RALT                Right MENU key
    BROWSER_BACK        Browser Back key
    BROWSER_FORWARD     Browser Forward key
    BROWSER_REFRESH     Browser Refresh key
    BROWSER_STOP        Browser Stop key
    BROWSER_SEARCH      Browser Search key
    BROWSER_FAVORITES   Browser Favorites key
    BROWSER_HOME        Browser Start and Home key
    VOLUME_MUTE         Volume Mute key
    VOLUME_DOWN         Volume Down key
    VOLUME_UP           Volume Up key
    MEDIA_NEXT_TRACK    Next Track key
    MEDIA_PREV_TRACK    Previous Track key
    MEDIA_STOP          Stop Media key
    MEDIA_PLAY_PAUSE    Play/Pause Media key
    LAUNCH_MAIL         Start Mail key
    LAUNCH_MEDIA_SELECT Select Media key
    LAUNCH_APP1         Start Application 1 key
    LAUNCH_APP2         Start Application 2 key
    OEM_1               Used for miscellaneous characters; it can vary by keyboard.For the US standard keyboard, the ';:' key
    OEM_PLUS            For any country/region, the '+' key
    OEM_COMMA           For any country/region, the ',' key
    OEM_MINUS           For any country/region, the '-' key
    OEM_PERIOD          For any country/region, the '.' key
    OEM_2               Used for miscellaneous characters; it can vary by keyboard.
    OEM_3               Used for miscellaneous characters; it can vary by keyboard.
    OEM_4               Used for miscellaneous characters; it can vary by keyboard.
    OEM_5               Used for miscellaneous characters; it can vary by keyboard.
    OEM_6               Used for miscellaneous characters; it can vary by keyboard.
    OEM_7               Used for miscellaneous characters; it can vary by keyboard.
    OEM_8               Used for miscellaneous characters; it can vary by keyboard.
    OEM_102             Either the angle bracket key or the backslash key on the RT 102-key keyboard
    PROCESSKEY          IME PROCESS key
    PACKET              Used to pass Unicode characters as if they were keystrokes. The VK_PACKET key is the low word of a 32-bit Virtual Key value used for non-keyboard input methods. For more information, see Remark in KEYBDINPUT, SendInput, WM_KEYDOWN, and WM_KeyUp
    ATTN                Attn key
    CRSEL               CrSel key
    EXSEL               ExSel key
    EREOF               Erase EOF key
    PLAY                Play key
    ZOOM                Zoom key
    NONAME              Reserved
    PA1                 PA1 key
    OEM_CLEAR           Clear key
    =================== =======================

    Examples.

    .. code-block:: python

        lib = Windows()
        # {Ctrl}, {Delete} ... are special keys' name in SpecialKeyNames.
        lib.send_keys('{Ctrl}a{Delete}{Ctrl}v{Ctrl}s{Ctrl}{Shift}s{Win}e{PageDown}') #press Ctrl+a, Delete, Ctrl+v, Ctrl+s, Ctrl+Shift+s, Win+e, PageDown
        lib.send_keys('{Ctrl}(AB)({Shift}(123))') #press Ctrl+A+B, type '(', press Shift+1+2+3, type ')', if '()' follows a hold key, hold key won't release util ')'
        lib.send_keys('{Ctrl}{a 3}') #press Ctrl+a at the same time, release Ctrl+a, then type 'a' 2 times
        lib.send_keys('{a 3}{B 5}') #type 'a' 3 times, type 'B' 5 times
        lib.send_keys('{{}Hello{}}abc {a}{b}{c} test{} 3}{!}{a} (){(}{)}') #type: '{Hello}abc abc test}}}!a ()()'
        lib.send_keys('0123456789{Enter}')
        lib.send_keys('ABCDEFGHIJKLMNOPQRSTUVWXYZ{Enter}')
        lib.send_keys('abcdefghijklmnopqrstuvwxyz{Enter}')
        lib.send_keys('`~!@#$%^&*()-_=+{Enter}')
        lib.send_keys('[]{{}{}}\\|;:\'\",<.>/?{Enter}')

    Using access key of the element (element property -> AccessKey 'alt+s').
    The `(+s)` means that previous special key is kept down until closing parenthesis is reached.

    On the below example this means that 'ALT' key is pressed down, then '+' and 's' keys are pressed
    down before they are all released up.

    .. code-block:: robotframework

        Send Keys   keys={Alt}(+s)

    Mouse clicks can be executed with keywords specific for a type of a click, e.g. ``Click`` (normal click),
    ``Double Click`` and ``Right Click``.

    **How to inspect**

    Most common, and recommended by Microsoft, inspector tool for Windows is `Accessibility Insights`_ that
    can be installed separately. Other options are tools `Inspect Object`_  and `UI Automation Verify`_, which
    can be accessed by installing Windows SDK.

    A more programmatic approach is to run `Print Tree    log_as_warnings=${True}`
    keyword and then observe in the logs the found elements structure starting from
    Desktop as root. (refer to keyword's documentation for more details)

    .. _Accessibility Insights: https://accessibilityinsights.io/
    .. _Inspect Object: https://docs.microsoft.com/en-us/windows/win32/winauto/inspect-objects
    .. _UI Automation Verify: https://docs.microsoft.com/en-us/windows/win32/winauto/ui-automation-verify

    **Recording**

    The package provides some rudimentary inspecting and recording via script ``windows-record``, which can
    be started in the command line (in a environment containing ``rpaframework-windows`` installation).

    Recording inspects elements on **mouse click** and can be stopped with keyboard **ESC**.
    Expected console output.

    .. code-block:: bash

        C:\\Users\\User\\robots\\>windows-record  # or >python -m RPA.Windows
        keyboard and mouse listeners started

        --------------------------------------------------------------------------------
        COPY & PASTE BELOW CODE INTO *** Tasks *** or *** Keywords ***
        --------------------------------------------------------------------------------

        Control Window    Taskbar  # Handle: 131380
        Click   name:"Type here to search"
        Control Window    Calculator  # Handle: 3411840
        Click   name:Five
        Click   name:Eight
        Click   name:Five

        --------------------------------------------------------------------------------


    **Examples**

    Both Robot Framework and Python examples follow.

    The library must be imported first.

    .. code-block:: robotframework

        *** Settings ***
        Library    RPA.Windows


    Windows Calculator task

    .. code-block:: robotframework

        *** Tasks ***
        Do some calculations
            [Setup]  Windows Run   calc.exe
            Control Window    name:Calculator
            Click    id:clearButton
            Send Keys   keys=96+4=
            ${result}=    Get Attribute    id:CalculatorResults    Name
            Log To Console    ${result}
            ${buttons}=  Get Elements  type:Group and name:"Number pad" > type:Button
            FOR  ${button}  IN  @{buttons}
                Log To Console   ${button}
            END
            [Teardown]   Close Current Window

    Python example

    .. code-block:: robotframework

        from RPA.Windows import Windows

        library = Windows()

        def test_do_some_calculations():
            try:
                library.windows_run("calc.exe")
                library.control_window("name:Calculator")
                library.click("id:clearButton")
                library.send_keys(keys="96+4=")
                result = library.get_attribute("id:CalculatorResults", "Name")
                print(result)
                buttons = library.get_elements('type:Group and name:"Number pad" > type:Button')
                for button in buttons:
                    print(button)
            finally:
                library.close_current_window()

    """  # noqa: E501,W605

    ROBOT_LIBRARY_SCOPE = "GLOBAL"
    ROBOT_LIBRARY_DOC_FORMAT = "REST"
    SIMULATE_MOVE = False

    def __init__(self, locators_path: Optional[str] = None):
        self.wait_time: float = 0.5

        # Prevent comtypes writing a lot of log messages.
        comtypes_logger = logging.getLogger("comtypes")
        comtypes_logger.propagate = False
        if utils.IS_WINDOWS:
            # Disable uiautomation writing into a log file.
            Logger.SetLogFile("")

        super().__init__(locators_path=locators_path)

    def _get_libraries(self, locators_path: Optional[str]):
        return [
            ActionKeywords(self),
            ElementKeywords(self),
            LocatorKeywords(self, locators_path=locators_path),
            WindowKeywords(self),
        ]



from robot.api.deco import keyword  # imported from here by keyword modules

from .action import ActionKeywords
from .context import (
    ActionNotPossible,
    ElementNotFound,
    LibraryContext,
    ControlNotFound,
    MultipleControlsFound,
    TimeoutException,
    WindowControlError,
    with_timeout,
)
from .elements import ElementKeywords
from .locators import Locator, LocatorKeywords
from .window import WindowKeywords



from datetime import datetime
from typing import List

import fire
from pynput_robocorp import mouse, keyboard  # pylint: disable=C0415
from RPA.core.windows.inspect import ElementInspector, RecordElement


recording_time = None
recording: List[RecordElement] = []


def stop_listeners():
    mouse.Listener.stop()
    keyboard.Listener.stop()


def start_recording(verbose: bool = False):
    """Start recording mouse clicks.

    Can be stopped by pressing keyboard ``ESC``.
    """
    global recording  # pylint: disable=W0603
    recording = []

    def on_click(x, y, button, pressed):  # pylint: disable=W0613
        global recording, recording_time  # pylint: disable=W0602
        if pressed:
            inspect_time = datetime.now()
            if recording_time:
                timediff = inspect_time - recording_time
                seconds = max(round(float(timediff.microseconds / 1000000.0), 1), 0.1)
                recording.append({"type": "sleep", "value": seconds})
            recording_time = inspect_time
            ElementInspector.inspect_element(recording=recording, verbose=verbose)

    def on_release(key):
        if key == keyboard.Key.esc:
            return False
        return True

    mouse_listener = mouse.Listener(on_click=on_click)
    mouse_listener.start()
    with keyboard.Listener(on_release=on_release) as key_listener:
        print("keyboard and mouse listeners started", flush=True)
        key_listener.join()
    mouse_listener.stop()
    the_recording = get_recording()
    print(the_recording)


def get_recording(sleeps: bool = False):
    """Get list of recorded steps.

    :param sleeps: Exclude recording sleeps when `False`.
    """
    # NOTE: Works with "Click" only for now.
    global recording  # pylint: disable=W0602
    output = []
    top = None
    action_name = "Click"
    for item in recording:
        if sleeps and item["type"] == "sleep":
            output.append(f"Sleep   {item['value']}s")
        if (
            item["type"] == "locator"
            and not top
            or "top" in item
            and item["top"] != top
        ):
            output.append(
                f"Control Window    {item['top']}  # Handle: {item['top_handle']}"
            )
            top = item["top"]
        if item["type"] == "locator":
            output.append(f"{action_name}   {item['locator']}")

    result = "\n".join(output)
    header = (
        f"\n{'-'*80}"
        "\nCOPY & PASTE BELOW CODE INTO *** Tasks *** or *** Keywords ***"
        f"\n{'-'*80}\n\n"
    )
    footer = f"\n\n{'-'*80}"
    return f"{header}{result}{footer}"


def main():
    fire.Fire(start_recording)



from typing import List, Optional

from RPA.core.windows.locators import (
    Locator,
    LocatorMethods,
    MatchObject,
    WindowsElement,
)

from RPA.Windows import utils
from RPA.Windows.keywords import keyword
from RPA.Windows.keywords.context import with_timeout

if utils.IS_WINDOWS:
    import uiautomation as auto
    from uiautomation import TreeNode


class LocatorKeywords(LocatorMethods):
    """Keywords for handling Windows locators."""

    # NOTE(cmin764): Timeout is automatically set to `None` in the upper calls by the
    #  `with_timeout` decorator, so we alter the behaviour (context timeout setting)
    #  on the first call only.

    @keyword
    @with_timeout
    def get_element(
        self,
        locator: Optional[Locator] = None,
        search_depth: int = 8,
        root_element: Optional[WindowsElement] = None,
        timeout: Optional[float] = None,
    ) -> WindowsElement:
        """Get a Control Windows element defined by the locator.

        The returned element can be used instead of a locator string for other keywords
        accepting the `locator` parameter.
        Keyword ``Get Attribute`` can be used to read element attribute values.

        If `locator` is `None`, then the returned element will be in this priority:

            1. `root_element` if provided.
            2. Anchor element if that has been previously set with ``Set Anchor``.
            3. Current active window if that has been set with ``Control Window``.
            4. Last resort is the "Desktop" element.

        :param locator: Locator as a string or as an element object.
        :param search_depth: How deep the element search will traverse. (default 8)
        :param root_element: Will be used as search root element object if provided.
        :param timeout: After how many seconds (float) to give up on search. (see
            keyword ``Set Global Timeout``)
        :returns: The identified `WindowsElement` object.

        **Example: Robot Framework**

        .. code-block:: robotframework

            *** Tasks ***
            Set Text Into Notepad Window
                Windows Run    Notepad
                Control Window      subname:Notepad
                ${element} =    Get Element    regex:"Text (E|e)ditor"
                Set Value    ${element}    note to myself
        """
        return super().get_element(
            locator=locator,
            search_depth=search_depth,
            root_element=root_element,
            timeout=timeout,
        )

    @classmethod
    def _search_siblings(
        cls, initial_element: WindowsElement, *, locator: Optional[Locator]
    ) -> List[WindowsElement]:
        elements = [initial_element]

        element = initial_element
        while True:
            next_control = element.item.GetNextSiblingControl()
            if not next_control:
                break

            element = WindowsElement(next_control, locator)
            if initial_element.is_sibling(element):
                # Every newly found matching element will inherit the offset set in the
                #  original initially found element.
                element.item.robocorp_click_offset = (
                    initial_element.item.robocorp_click_offset
                )
                elements.append(element)

        return elements

    def _search_globally(
        self,
        initial_element: WindowsElement,
        *,
        locator: Optional[Locator],
        search_depth: int,
        root_element: Optional[WindowsElement],
        timeout: Optional[float],
    ) -> List[WindowsElement]:
        def get_first_child(ctrl: TreeNode) -> TreeNode:
            return ctrl.GetFirstChildControl()

        def get_next_sibling(ctrl: TreeNode) -> TreeNode:
            return ctrl.GetNextSiblingControl()

        def compare(ctrl: TreeNode, _) -> bool:
            element = WindowsElement(ctrl, locator)
            return initial_element.is_sibling(element)

        # Take all the elements (no matter their level) starting from a parent as
        #  search tree root.
        parent_locator = None
        if locator:
            branches = locator.rsplit(MatchObject.TREE_SEP, 1)
            if len(branches) == 2:
                # Full locator's parent becomes the root for the search.
                parent_locator = branches[0]
        top_element = self.get_element(
            # If the locator doesn't have a parent (null `parent_locator`) then simply
            #  rely on the resulting root resolution: root > anchor > window > Desktop.
            parent_locator,
            search_depth,
            root_element,
            timeout,
        )
        # Explore the entire subtree of elements starting with the resulted root above
        #  and keep only the ones matching the strategies in the last locator branch.
        tree_generator = auto.WalkTree(
            top_element.item,
            getFirstChild=get_first_child,
            getNextSibling=get_next_sibling,
            yieldCondition=compare,
            includeTop=True,
            maxDepth=search_depth,
        )
        elements = []
        for control, _ in tree_generator:
            element = WindowsElement(control, locator)
            # Every newly found matching element will inherit the offset set in the
            #  original initially found element.
            element.item.robocorp_click_offset = (
                initial_element.item.robocorp_click_offset
            )
            elements.append(element)
        return elements

    @keyword
    @with_timeout
    def get_elements(
        self,
        locator: Optional[Locator] = None,
        search_depth: int = 8,
        root_element: Optional[WindowsElement] = None,
        timeout: Optional[float] = None,
        siblings_only: bool = True,
    ) -> List[WindowsElement]:
        """Get a list of elements matching the locator.

        By default, only the siblings (similar elements on the same level) are taken
        into account. In order to search globally, turn `siblings_only` off, but be
        aware that this will take more time to process.
        For more details on the rest of parameters, take a look at the ``Get Element``
        keyword.

        :param locator: Locator as a string or as an element object.
        :param search_depth: How deep the element search will traverse. (default 8)
        :param root_element: Will be used as search root element object if provided.
        :param timeout: After how many seconds (float) to give up on search. (see
            keyword ``Set Global Timeout``)
        :param siblings_only: Filter for elements on the same level as the initially
            found one. Turn it off for a global search. (`True` by default)
        :returns: A list of matching `WindowsElement` objects.

        **Example: Robot Framework**

        .. code-block:: robotframework

            *** Tasks ***
            Get Headers On Same Level
                Set Anchor      id:DataGrid
                @{elements} =   Get Elements    type:HeaderItem
                FOR    ${element}    IN    @{elements}
                    Log To Console    ${element.name}
                END

            Get All Calculator Buttons Matching Expression
                Windows Run    Calc
                Control Window    subname:Calc
                @{buttons} =    Get Elements    class:Button regex:.*o.*
                ...     siblings_only=${False}
                Log List    ${buttons}
                ${length} =     Get Length      ${buttons}
                Log To Console      Number of buttons: ${length}
        """
        initial_element = self.get_element(locator, search_depth, root_element, timeout)

        if siblings_only:
            return self._search_siblings(initial_element, locator=locator)

        return self._search_globally(
            initial_element,
            locator=locator,
            search_depth=search_depth,
            root_element=root_element,
            timeout=timeout,
        )



import platform

# pylint: disable=unused-import
from RPA.core.windows.helpers import (  # noqa: F401
    IS_WINDOWS,
    get_process_list,
    is_numeric,
)


def call_attribute_if_available(object_name, attribute_name):
    if hasattr(object_name, attribute_name):
        return getattr(object_name, attribute_name)()
    return None


def get_win_version() -> str:
    """Windows only utility which returns the current Windows major version."""
    # Windows terminal `ver` command is bugged, until that's fixed, check by build
    #  number. (the same applies for `platform.version()`)
    version_parts = platform.version().split(".")
    major = version_parts[0]
    if major == "10" and int(version_parts[2]) >= 22000:
        major = "11"

    return major



import os
import signal
import time
from typing import Dict, List, Optional

from RPA.core.windows.locators import Locator, WindowsElement
from RPA.core.windows.window import WindowMethods

from RPA.Windows import utils
from RPA.Windows.keywords import (
    ElementNotFound,
    WindowControlError,
    keyword,
    with_timeout,
)

if utils.IS_WINDOWS:
    import uiautomation as auto


class WindowKeywords(WindowMethods):
    """Keywords for handling Window controls"""

    @staticmethod
    def _iter_locator(locator: Optional[Locator]) -> Optional[Locator]:
        if not locator:
            yield locator  # usually `None`
        elif isinstance(locator, WindowsElement):
            yield locator  # yields element as it is
        elif "type:" in locator or "control:" in locator:
            yield locator  # yields rigid string locator
        else:
            # yields flexible string locators with different types
            yield f"{locator} and type:WindowControl"
            yield f"{locator} and type:PaneControl"

    @keyword(tags=["window"])
    @with_timeout
    def control_window(
        self,
        locator: Optional[Locator] = None,
        foreground: bool = True,
        wait_time: Optional[float] = None,
        timeout: Optional[float] = None,  # pylint: disable=unused-argument
        main: bool = True,
    ) -> WindowsElement:
        """Controls the window defined by the locator.

        This means that this window is used as a root element
        for all the following keywords using locators.

        Returns `WindowsElement`.

        :param locator: string locator or Control element
        :param foreground: True to bring window to foreground
        :param wait_time: time to wait after activating a window
        :param timeout: float value in seconds, see keyword
         ``Set Global Timeout``
        :param main: on True (default) starts search from desktop level,
         on False will continue search on child elements of current
         active window
        :return: WindowsElement object

        Example:

        .. code-block:: robotframework

            Control Window   Calculator
            Control Window   name:Calculator
            Control Window   subname:Notepad
            Control Window   regex:.*Notepad
            ${window}=  Control Window   executable:Spotify.exe
        """
        for loc in self._iter_locator(locator):
            self.ctx.window_element = self._find_window(
                loc, main
            )  # works with windows too
            if self.ctx.window_element:
                break  # first window found is enough

        window = self.window
        if window is None:
            raise WindowControlError(
                f"Could not locate window with locator: {locator!r} "
                f"(timeout: {self.current_timeout})"
            )

        if foreground:
            self.foreground_window()
        if wait_time:
            time.sleep(wait_time)
        return window

    @keyword(tags=["window"])
    def control_child_window(
        self,
        locator: Optional[Locator] = None,
        foreground: bool = True,
        wait_time: Optional[float] = None,
        timeout: Optional[float] = None,
    ) -> WindowsElement:
        """Get control of child window of the active window
        by locator.

        :param locator: string locator or Control element
        :param foreground: True to bring window to foreground
        :param wait_time: time to wait after activeting a window
        :param timeout: float value in seconds, see keyword
         ``Set Global Timeout``
        :return: WindowsElement object

        Example:

        .. code-block:: robotframework

            Control Window   subname:"Sage 50" type:Window
            # actions on the main application window
            # ...
            # get control of child window of Sage application
            Control Child Window   subname:"Test Company" depth:1
        """
        return self.control_window(locator, foreground, wait_time, timeout, main=False)

    def _find_window(self, locator, main) -> Optional[WindowsElement]:
        try:
            # `root_element = None` means using the `anchor` or `window` as root later
            #  on. (fallbacks to Desktop)
            root_element = (
                WindowsElement(auto.GetRootControl(), locator) if main else None
            )
            window = self.ctx.get_element(locator, root_element=root_element)
            return window
        except (ElementNotFound, LookupError):
            return None

    @keyword(tags=["window"])
    def foreground_window(self, locator: Optional[Locator] = None) -> WindowsElement:
        """Bring the current active window or the window defined
        by the locator to the foreground.

        :param locator: string locator or Control element
        :return: WindowsElement object

        Example:

        .. code-block:: robotframework

            ${window}=  Foreground Window   Calculator
        """
        if locator:
            return self.control_window(locator, foreground=True)
        window = self.window
        if window is None:
            raise WindowControlError("There is no active window")

        utils.call_attribute_if_available(window.item, "SetFocus")
        utils.call_attribute_if_available(window.item, "SetActive")
        window.item.MoveCursorToMyCenter(simulateMove=self.ctx.SIMULATE_MOVE)
        return window

    def _resize_window(
        self, locator: Optional[Locator] = None, *, attribute: str
    ) -> WindowsElement:
        if locator:
            self.control_window(locator)
        window = self.window
        if window is None:
            raise WindowControlError("There is no active window")

        attr_func = getattr(window.item, attribute, None)
        if attr_func:
            attr_func()
        else:
            self.logger.warning(
                "Element %r does not have the %r attribute", window, attribute
            )
        return window

    @keyword(tags=["window"])
    def minimize_window(self, locator: Optional[Locator] = None) -> WindowsElement:
        """Minimize the current active window or the window defined
        by the locator.

        :param locator: string locator or element
        :return: `WindowsElement` object

        Example:

        .. code-block:: robotframework

            ${window} =    Minimize Window  # Current active window
            Minimize Window    executable:Spotify.exe
        """
        return self._resize_window(locator, attribute="Minimize")

    @keyword(tags=["window"])
    def maximize_window(self, locator: Optional[Locator] = None) -> WindowsElement:
        """Maximize the current active window or the window defined
        by the locator.

        :param locator: string locator or element
        :return: `WindowsElement` object

        Example:

        .. code-block:: robotframework

            ${window} =    Maximize Window  # Current active window
            Maximize Window    executable:Spotify.exe
        """
        return self._resize_window(locator, attribute="Maximize")

    @keyword(tags=["window"])
    def restore_window(self, locator: Optional[Locator] = None) -> WindowsElement:
        """Window restore the current active window or the window
        defined by the locator.

        :param locator: string locator or element
        :return: `WindowsElement` object

        Example:

        .. code-block:: robotframework

            ${window} =    Restore Window  # Current active window
            Restore Window    executable:Spotify.exe
        """
        return self._resize_window(locator, attribute="Restore")

    @keyword(tags=["window"])
    def list_windows(
        self, icons: bool = False, icon_save_directory: Optional[str] = None
    ) -> List[Dict]:
        """List all window element on the system.

        :param icons: on True dictionary will contain Base64
         string of the icon, default False
        :param icon_save_directory: if set will save retrieved icons
         into this filepath, by default icon files are not saved
        :return: list of dictionaries containing information
         about Window elements

        Example:

        .. code-block:: robotframework

            ${windows}=  List Windows
            FOR  ${window}  IN  @{windows}
                Log  Window title:${window}[title]
                Log  Window process name:${window}[name]
                Log  Window process id:${window}[pid]
                Log  Window process handle:${window}[handle]
            END
        """
        return super().list_windows(
            icons=icons, icon_save_directory=icon_save_directory
        )

    @keyword(tags=["window"])
    def windows_run(self, text: str, wait_time: float = 3.0) -> None:
        """Use Windows run window to launch application.

        Activated by pressing `win + r`.

        :param text: text to enter into run input field
        :param wait_time: sleep time after search has been entered (default 3.0 seconds)

        Example:

        .. code-block:: robotframework

            Windows Run   explorer.exe
        """
        self.ctx.send_keys(None, "{Win}r")
        self.ctx.send_keys(None, text)
        self.ctx.send_keys(None, "{Enter}")
        time.sleep(wait_time)

    @keyword(tags=["window"])
    def windows_search(self, text: str, wait_time: float = 3.0) -> None:
        """Use Windows search window to launch application.

        Activated by pressing `win + s`.

        :param text: text to enter into search input field
        :param wait_time: sleep time after search has been entered (default 3.0 seconds)

        Example:

        .. code-block:: robotframework

            Windows Search   Outlook
        """
        search_cmd = "{Win}s"
        if utils.get_win_version() == "11":
            search_cmd = search_cmd.rstrip("s")
        self.ctx.send_keys(None, search_cmd)
        self.ctx.send_keys(None, text)
        self.ctx.send_keys(None, "{Enter}")
        time.sleep(wait_time)

    @keyword(tags=["window"])
    def close_current_window(self) -> bool:
        """Closes current active window or logs a warning message.

        :return: True if close was successful, False if not

        Example:

        .. code-block:: robotframework

            ${status}=  Close Current Window
        """
        window = self.window
        if window is None:
            self.logger.warning("There is no active window!")
            self.ctx.window_element = None
            return False

        pid = window.item.ProcessId
        self.logger.info("Closing window with name: %s (PID: %d)", window.name, pid)
        os.kill(pid, signal.SIGTERM)
        self.ctx.window_element = None

        anchor = self.ctx.anchor_element
        if anchor and window.is_sibling(anchor):
            # We just closed the anchor as well (along with its relatives), so clear
            #  it out properly.
            self.ctx.clear_anchor()

        return True

    @keyword(tags=["window"])
    @with_timeout
    def close_window(
        self,
        locator: Optional[Locator] = None,
        timeout: Optional[float] = None,  # pylint: disable=unused-argument
    ) -> int:
        """Closes identified windows or logs the problems.

        :param locator: String locator or `Control` element.
        :param timeout: float value in seconds, see keyword
         ``Set Global Timeout``
        :return: How many windows were found and closed.

        Example:

        .. code-block:: robotframework

            ${closed_count} =     Close Window    Calculator
        """
        # Starts the search from Desktop level.
        root_element = WindowsElement(auto.GetRootControl(), locator)
        # With all flavors of locators. (if flexible)
        for loc in self._iter_locator(locator):
            try:
                elements: List[WindowsElement] = self.ctx.get_elements(
                    loc, root_element=root_element
                )
            except (ElementNotFound, LookupError):
                continue
            break
        else:
            raise WindowControlError(f"Couldn't find any window with {locator!r}")

        closed = 0
        for element in elements:
            self.logger.debug("Controlling and closing window: %s", element)
            try:
                self.control_window(element)
                closed += int(self.close_current_window())
            except Exception as exc:  # pylint: disable=broad-except
                self.logger.warning("Couldn't close window %r due to: %s", element, exc)
        return closed

    @keyword(tags=["window"])
    def get_os_version(self) -> str:
        """Returns the current Windows major version as string.

        Example:

        .. code-block:: robotframework

            ${ver} =     Get OS Version
            Log     ${ver}  # 10
        """
        return utils.get_win_version()



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['RPA', 'RPA.Windows', 'RPA.Windows.keywords', 'RPA.scripts']

package_data = \
{'': ['*']}

install_requires = \
['fire>=0.4.0,<0.5.0',
 'pynput-robocorp-fork>=4.0.0,<5.0.0',
 'robotframework-pythonlibcore>=3.0.0,<4.0.0',
 'robotframework>=4.0.0,!=4.0.1,<6.0.0',
 'rpaframework-core>=10.0.4,<11.0.0',
 'uiautomation>=2.0.15,<3.0.0']

extras_require = \
{':python_full_version != "3.7.6" and python_full_version != "3.8.1" and sys_platform == "win32"': ['pywin32>=300,<304'],
 ':python_version >= "3.6" and python_version < "3.7"': ['dataclasses>=0.7,<0.8'],
 ':sys_platform == "win32"': ['comtypes==1.1.11', 'psutil>=5.9.0,<6.0.0']}

entry_points = \
{'console_scripts': ['windows-record = RPA.scripts.record:main']}

setup_kwargs = {
    'name': 'rpaframework-windows',
    'version': '6.2.0',
    'description': 'Windows library for RPA Framework',
    'long_description': 'rpaframework-windows\n====================\n\nThis library enables Windows automation for the  `RPA Framework`_\nlibraries based on `uiautomation`_ dependency.\n\n.. _RPA Framework: https://rpaframework.org\n.. _uiautomation: https://github.com/yinkaisheng/Python-UIAutomation-for-Windows\n',
    'author': 'RPA Framework',
    'author_email': 'rpafw@robocorp.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://rpaframework.org/',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
